from ..proposal.clustering import ClusteringFlowProposal
from ...gw.proposal import GWFlowProposal


class ClusteringGWFlowProposal(ClusteringFlowProposal, GWFlowProposal):
    """Augmented version of GWFlowProposal.

    See :obj:`~nessai.proposal.augmented.ClusteringFlowProposal` and
    :obj:`~nessai.gw.proposal.GWFlowPropsosal`
    """

    pass
